/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.openstack.swift.storage;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.URI;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Locale;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.encryption.Encryption;
import org.dasein.cloud.jclouds.openstack.swift.Swift;
import org.dasein.cloud.storage.AbstractBlobStoreSupport;
import org.dasein.cloud.storage.CloudStoreObject;
import org.dasein.cloud.storage.FileTransfer;
import org.dasein.util.Jiterator;
import org.dasein.util.JiteratorPopulator;
import org.dasein.util.PopulatorThread;
import org.jclouds.blobstore.BlobStoreContext;
import org.jclouds.blobstore.domain.Blob;
import org.jclouds.blobstore.domain.PageSet;
import org.jclouds.blobstore.domain.StorageMetadata;
import org.jclouds.blobstore.domain.StorageType;

public class SwiftBlobSupport extends AbstractBlobStoreSupport {
    static private final Logger logger = Logger.getLogger(SwiftBlobSupport.class);
    
    static public final String SEPARATOR = ".";
    
    static private String toAbstract(String nameFromGoogle) {
        return nameFromGoogle.replace(SEPARATOR, ".");
    }
    
    static private final String toSwift(String abstractName) {
        return abstractName.replaceAll("\\.", SEPARATOR);
    }
    
    private Swift cloud;
    
    SwiftBlobSupport(Swift cloud) { this.cloud = cloud; }

    public void clear(String directoryName) throws CloudException, InternalException {
        logger.debug("enter - clear(String)");
        try {
            for( CloudStoreObject item : listFiles(directoryName) ) {
                if( item.isContainer() ) {
                    clear(item.getDirectory() + "." + item.getName());
                }
                else {
                    removeFile(directoryName, item.getName(), false);
                }
            }
            BlobStoreContext ctx = cloud.getSwiftContext();

            try {
                getClient(ctx).deleteContainer(toSwift(directoryName));
            }
            finally {
                ctx.close();
            }
        }
        finally {
            logger.debug("exit - clear(String)");
        }
    }

    public String createDirectory(String abstractDirectoryName, boolean findFreeName) throws InternalException, CloudException {
        logger.debug("enter - createDirectory(String, boolean)");
        try {
            try {
                String[] path = abstractDirectoryName.split("\\.");
                
                if( path == null || path.length < 1 ) {
                    path = new String[] { abstractDirectoryName };
                }
                for( int i=0; i<path.length; i++ ) {
                    String root = null;
                    
                    path[i] = verifyName(path[i], true);
                    if( i > 0 ) {
                        StringBuilder str = new StringBuilder();
                        
                        for( int j=0; j<i; j++ ) {
                            if( j > 0 ) {
                                str.append(".");
                            }
                            str.append(path[j]);
                        }
                        root = str.toString();
                    }
                    String p;
                    
                    if( root == null ) {
                        p = path[i];
                    }
                    else {
                        p = root + "." + path[1];
                    }
                    if( !exists(p) ) {
                        createDirectory(root, path[i]);
                    }
                    else if( i == path.length-1 ) {
                        if( !findFreeName ) {
                            throw new CloudException("The directory " + abstractDirectoryName + " already exists.");
                        }
                        else {
                            String tempName = path[i];
                            String suffix = "-";
                            char c = 'a';
                            
                            while( true ) {
                                path[i] = tempName + suffix + c;
                                if( root == null ) {
                                    p = path[i];
                                }
                                else {
                                    p = root + "." + path[1];
                                }
                                if( exists(p) ) {
                                    break;
                                }
                                if( c == 'z' ) {
                                    suffix = suffix + "a";
                                    c = 'a';
                                }
                                else {
                                    c++;
                                }
                            }
                            createDirectory(root, path[i]);
                        }
                    }
                }
                return join(".", path);
            } 
            catch( CloudException e ) {
                logger.error(e);
                e.printStackTrace();
                throw e;
            } 
            catch(InternalException e ) {
                logger.error(e);
                e.printStackTrace();
                throw e;
            } 
            catch( RuntimeException e ) {
                logger.error(e);
                e.printStackTrace();
                throw new InternalException(e);
            }
            catch( Error e ) {
                logger.error(e);
                e.printStackTrace();
                throw new InternalException(e);
            }
        }
        finally {
            logger.debug("exit - createDirectory(String, boolean)");
        }
    }
    
    @Override
    public long exists(String abstractDirectoryName, String object, boolean multiPart) throws InternalException, CloudException {
        logger.debug("enter - exists(String, String, boolean)");
        try {
            if( !multiPart ) {
                try {
                    BlobStoreContext ctx = cloud.getSwiftContext();
                
                    try {
                        if( object == null ) {
                            int idx = abstractDirectoryName.lastIndexOf('.');
                            String dir;
                            
                            if( idx == -1 ) {
                                dir = null;
                            }
                            else {
                                dir = abstractDirectoryName.substring(0, idx);
                            }
                            Collection<StorageMetadata> containers = filter(getClient(ctx).list(), dir);
                            
                            for( StorageMetadata container : containers ) {
                                if( toAbstract(container.getName()).equals(abstractDirectoryName) ) {
                                    return 0L;
                                }
                            }
                            return -1L;
                        }
                        else {
                                Blob blob = getClient(ctx).getBlob(toSwift(abstractDirectoryName), object);
                                Number len;
                                
                                if( blob == null ) {
                                    return -1L;
                                }
                                len = blob.getMetadata().getContentMetadata().getContentLength();
                                return (len == null ? -1L : len.longValue());
                        }
                    }
                    finally {
                        ctx.close();
                    }
                }
                catch( RuntimeException e ) {
                    logger.error("Could not retrieve file info for " + abstractDirectoryName + "." + object + ": " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
            }
            else {
                if( exists(abstractDirectoryName, object + ".properties", false) == -1L ) {
                    return -1L;
                }
                Properties properties = new Properties();
                String str;
                
                try {
                    BlobStoreContext ctx = cloud.getSwiftContext();

                    try {
                        Blob blob = getClient(ctx).getBlob(toSwift(abstractDirectoryName), object + ".properties");
                        
                        if( blob == null ) {
                            throw new CloudException("File was modified while we were reading it.");
                        }
                        try {
                            properties.load(blob.getPayload().getInput());
                        }
                        catch( IOException e ) {
                            logger.error("IO error loading file data for " + abstractDirectoryName + "." + object + ": " + e.getMessage());
                            e.printStackTrace();
                            throw new InternalException(e);
                        }
                    }
                    finally {
                        ctx.close();
                    }
                }
                catch( RuntimeException e ) {
                    logger.error("Could not retrieve file info for " + abstractDirectoryName + "." + object + ": " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
                str = properties.getProperty("length");
                if( str == null ) {
                    return 0L;
                }
                else {
                    return Long.parseLong(str);
                }
            }
        }
        finally {
            logger.debug("exit - exists(String, String, boolean)");
        }
    }
    
    private org.jclouds.blobstore.BlobStore getClient(BlobStoreContext ctx) {
        return ctx.getBlobStore();
    }
    
    @Override
    public long getMaxFileSizeInBytes() {
        return 5000000000L;
    }

    @Override
    public String getProviderTermForDirectory(Locale locale) {
        return "container";
    }

    @Override
    public String getProviderTermForFile(Locale locale) {
        return "object";
    }

    @Override
    public String getSeparator() {
        return SEPARATOR;
    }
    
    @Override
    public Iterable<CloudStoreObject> listFiles(String parentDirectory) throws CloudException, InternalException {
        logger.debug("enter - listFiles(String)");
        try {
            PopulatorThread <CloudStoreObject> populator;
            final String dir = parentDirectory;
            
            cloud.hold();
            populator = new PopulatorThread<CloudStoreObject>(new JiteratorPopulator<CloudStoreObject>() {
                public void populate(Jiterator<CloudStoreObject> iterator) throws CloudException, InternalException {
                    try {
                        listFiles(dir, iterator);
                    }
                    finally {
                        cloud.release();
                    }
                }
            });
            populator.populate();
            return populator.getResult();
        }
        finally {
            logger.debug("exit - listFiles(String)");
        }
    }

    private void listFiles(String abstractDirectoryName, Jiterator<CloudStoreObject> iterator) throws CloudException, InternalException {
        loadDirectories(abstractDirectoryName, iterator);
        if( abstractDirectoryName != null ) {
            loadFiles(abstractDirectoryName, iterator);            
        }
    }
    
    private Collection<StorageMetadata> filter(PageSet<? extends StorageMetadata> pageSet, String prefix) {
        ArrayList<StorageMetadata> filtered = new ArrayList<StorageMetadata>();
        
        for( StorageMetadata container : pageSet ) {
            if( container.getType().equals(StorageType.CONTAINER) ) {
                String name = container.getName();
                
                if( name == null && prefix == null ) {
                    filtered.add(container);
                }
                else if( prefix == null ) {
                    if( name.indexOf(SEPARATOR) == -1 ) {
                        filtered.add(container);
                    }
                }
                else if( name != null && name.startsWith(prefix) && !name.equals(prefix) ) {
                    name = name.substring(prefix.length() + SEPARATOR.length());
                    if( name.indexOf(SEPARATOR) == -1 ) {
                        filtered.add(container);
                    }
                }
            }
        }
        return filtered;
    }
    
    private Collection<StorageMetadata> getBlobs(PageSet<? extends StorageMetadata> pageSet) {
        ArrayList<StorageMetadata> filtered = new ArrayList<StorageMetadata>();

        for( StorageMetadata container : pageSet ) {
            if( container.getType().equals(StorageType.BLOB) ) {
                filtered.add(container);
            }
        }
        return filtered;
    }
    
    private void loadDirectories(String abstractDirectoryName, Jiterator<CloudStoreObject> iterator) throws CloudException, InternalException {
        Collection<StorageMetadata> containers;
        
        try {
            BlobStoreContext ctx = cloud.getSwiftContext();

            try {
                if( abstractDirectoryName != null ) {
                    containers = filter(getClient(ctx).list(), toSwift(abstractDirectoryName) + SEPARATOR);
                }
                else {
                    containers = filter(getClient(ctx).list(), null);
                }
            }
            finally {
                ctx.close();
            }
        }
        catch( RuntimeException e ) {
            logger.error("Could not load directories in " + abstractDirectoryName + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException(e);
        }
        for( StorageMetadata container : containers ) {
            String name = container.getName();
            String azureDirectoryName;
            
            azureDirectoryName = (abstractDirectoryName == null ? null : toSwift(abstractDirectoryName));
            if( azureDirectoryName == null ) {
                if( name.indexOf(".") > -1 ) { // a sub-directory of another directory
                    continue;
                }
            }
            else if( name.equals(azureDirectoryName) ) {
                continue;
            }
            else if( !name.startsWith(azureDirectoryName + SEPARATOR) ) {
                continue;
            }
            if( azureDirectoryName != null ) {
                String tmp = name.substring(azureDirectoryName.length() + 1);
                int idx = tmp.indexOf(SEPARATOR);
                
                if( idx > 0 /* yes 0, not -1 */ && idx < (tmp.length()-SEPARATOR.length()) ) { // this is a sub of a sub
                    continue;
                }
            }
            CloudStoreObject file = new CloudStoreObject();
            name = toAbstract(name);
            String[] parts = name.split("\\.");
            
            if( parts == null || parts.length < 2 ) {
                file.setName(name);
                file.setDirectory(null);
            }
            else {
                StringBuilder dirName = new StringBuilder();
                
                file.setName(parts[parts.length-1]);
                for( int part=0; part<parts.length-1; part++ ) {
                    if( dirName.length() > 0 ) {
                        dirName.append(".");
                    }
                    dirName.append(parts[part]);
                }
                file.setDirectory(dirName.toString());
            }
            file.setContainer(true);
            file.setProviderRegionId(cloud.getContext().getRegionId());
            file.setSize(0L);
            file.setCreationDate(new Date());
            iterator.push(file);
        }
    }
        
    private void loadFiles(String abstractDirectoryName, Jiterator<CloudStoreObject> iterator) throws CloudException, InternalException {
        Collection<StorageMetadata> files;
        
        BlobStoreContext ctx = cloud.getSwiftContext();
        
        try {

            try {
                if( abstractDirectoryName == null ) {
                    files = getBlobs(getClient(ctx).list());
                }
                else {
                    files = getBlobs(getClient(ctx).list(toSwift(abstractDirectoryName)));
                }
            }
            finally {
                ctx.close();
            }
        }
        catch( RuntimeException e ) {
            logger.error("Could not list files in " + abstractDirectoryName + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException(e);
        }
        for( StorageMetadata info : files ) {
            CloudStoreObject file = new CloudStoreObject();
            Blob blob = getClient(ctx).getBlob(toSwift(abstractDirectoryName), info.getName());

            file.setContainer(false);
            file.setDirectory(abstractDirectoryName);
            file.setName(info.getName());
            file.setProviderRegionId(cloud.getContext().getRegionId());
            file.setSize(blob.getMetadata().getContentMetadata().getContentLength());
            URI uri = info.getUri();
            file.setLocation(uri == null ? null : uri.toString());
            file.setCreationDate(info.getLastModified());
            iterator.push(file);
        }
    }
    
    public void makePublic(String abstractFileName) throws InternalException, CloudException {
        throw new OperationNotSupportedException();         
    }

    public void makePublic(String abstractDirectoryName, String fileName) throws InternalException, CloudException {
        throw new OperationNotSupportedException();         
    }

    public void moveFile(String fromDirectory, String fileName, String targetRegionId, String toDirectory) throws InternalException, CloudException {
        logger.debug("enter - moveFile(String, String, String, String)");
        try {
            moveFile(fromDirectory, fileName, toDirectory);
        }
        finally {
            logger.debug("exit - moveFile(String, String, String, String)");
        }
    }


    @Override
    public void moveFile(String sourceDirectory, String object, String toDirectory) throws InternalException, CloudException {
        logger.debug("enter - moveFile(String, String, String)");
        try {
            CloudStoreObject directory = new CloudStoreObject();
            CloudStoreObject file = new CloudStoreObject();
            String[] parts = toDirectory.split("\\.");
            String dirPath, dirName;
            
            if( parts == null || parts.length < 2 ) {
                dirPath = null;
                dirName = toDirectory;
            }
            else {
                StringBuilder str = new StringBuilder();
                
                dirName = parts[parts.length-1];
                for( int i = 0; i<parts.length-1; i++ ) {
                    if( i > 0 ) {
                        str.append(".");
                    }
                    str.append(parts[i]);
                }
                dirPath = str.toString();
            }
            file.setContainer(false);
            file.setDirectory(sourceDirectory);
            file.setName(object);
            file.setProviderRegionId(cloud.getContext().getRegionId());
            directory.setContainer(true);
            directory.setDirectory(dirPath);
            directory.setName(dirName);
            directory.setProviderRegionId(cloud.getContext().getRegionId());
            copy(file, directory, object);
            removeFile(sourceDirectory, object, false); 
        }
        finally {
            logger.debug("exit - moveFile(String, String, String)");
        }
    }
    
    public void removeDirectory(String directory) throws CloudException, InternalException {
        logger.debug("enter - removeDirectory(String)");
        try {
            BlobStoreContext ctx = cloud.getSwiftContext();

            try {
                getClient(ctx).deleteContainer(toSwift(directory));
            }
            finally {
                ctx.close();
            }
        }
        finally {
            logger.debug("exit - removeDirectory(String)");
        }
    }

    public void removeFile(String directory, String name, boolean multipartFile) throws CloudException, InternalException {     
        logger.debug("enter - removeFile(String, String, boolean)");
        try {
            if( !multipartFile ) {
                removeFile(directory, name);
            }
            else {
                removeMultipart(directory, name);
            }
        }
        finally {
            logger.debug("exit - removeFile(String, String, boolean)");         
        }
    }

    @Override
    protected void removeFile(String directory, String name) throws CloudException, InternalException {
        logger.debug("enter - removeFile(String, String)");
        try {
            try {
                BlobStoreContext ctx = cloud.getSwiftContext();

                try {
                    getClient(ctx).removeBlob(toSwift(directory), name);
                }
                finally {
                    ctx.close();
                }
            }
            catch( RuntimeException e ) {
                logger.error("Failed to remove file " + directory + "." + name + ": " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            logger.debug("exit - removeFile(String,String)");
        }
    }
    
    public String renameDirectory(String oldName, String newName, boolean findFreeName) throws CloudException, InternalException {
        logger.debug("enter - renameDirectory(String, String, boolean)");
        try {
            String nd = createDirectory(newName, findFreeName);
            
            // list all the old files, move them
            for( CloudStoreObject f : listFiles(oldName) ) {
                moveFile(oldName, f.getName(), null, nd);
            }
            // delete the old container/objects
            removeDirectory(oldName);       
            return nd;
        }
        finally {
            logger.debug("exit - renameDirectory(String, String, boolean)");
        }
    }

    public void renameFile(String directory, String oldName, String newName) throws CloudException, InternalException {
        File tmp = null;
        
        try {
            // download old file
            tmp = File.createTempFile(newName, "tmp");          
            get(directory, oldName, tmp, null);
                    
            // create new name + upload file
            upload(tmp, directory, newName, false, null);
            
            // delete old name
            removeFile(directory, oldName, false);
        } 
        catch( CloudException e ) {
            logger.error(e);
            e.printStackTrace();
            throw e;
        } 
        catch( IOException e ) {
            logger.error(e);
            e.printStackTrace();
            throw new CloudException(e);
        } 
        catch( RuntimeException e ) {
            logger.error(e);
            e.printStackTrace();
            throw new InternalException(e);
        }
        finally {
            if (tmp != null) {
                tmp.delete();
            }
        }
    }
    
    public void upload(File sourceFile, String directory, String fileName, boolean multipart, Encryption encryption) throws CloudException, InternalException {
        try {
            if( multipart ) {
                try {
                    uploadMultipartFile(sourceFile, directory, fileName, encryption);
                }
                catch( InterruptedException e ) {
                    logger.error(e);
                    e.printStackTrace();
                    throw new CloudException(e.getMessage());
                }
            }
            else {
                try {
                    put(directory, fileName, sourceFile);
                }
                catch( NoSuchAlgorithmException e ) {
                    logger.error(e);
                    e.printStackTrace();
                    throw new InternalException(e);
                }
                catch( IOException e ) {
                    logger.error(e);
                    e.printStackTrace();
                    throw new CloudException(e.getMessage());
                }
            }        
        }
        finally {
            if( encryption != null ) {
                encryption.clear();
            }
        }
    }
    
    private boolean createDirectory(String parent, String name) throws InternalException, CloudException {
        logger.debug("enter - createDirectory(String)");
        try {
            try {
                BlobStoreContext ctx = cloud.getSwiftContext();

                try {
                    if( parent == null ) {
                        return getClient(ctx).createContainerInLocation(null, name);
                    }
                    else {
                        return getClient(ctx).createContainerInLocation(null, toSwift(parent + "." + name));
                    }
                }
                finally {
                    ctx.close();
                }
            }
            catch( RuntimeException e ) {
                logger.error("Could not create directory: " + e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            logger.debug("exit - createDirectory(String)");
        }
    }
    
    @Override
    protected void get(String directory, String location, File toFile, FileTransfer transfer) throws IOException, CloudException {
        logger.debug("enter - get(String, String, File, FileTransfer)");
        try {
            if( toFile.exists() ) {
                if( !toFile.delete() ) {
                    throw new IOException("File already exists that cannot be overwritten.");
                }
            }
            BlobStoreContext ctx = cloud.getSwiftContext();

            try {
                Blob blob = getClient(ctx).getBlob(toSwift(directory), location);
                
                if( blob == null ) {
                    throw new IOException("No such file: " + directory + "." + location);
                }
                copy(blob.getPayload().getInput(), new FileOutputStream(toFile), transfer);
            }
            finally {
                ctx.close();
            }
        }
        finally {
            logger.debug("exit - get(String, String, File, FileTransfer)");
        }
    }  
    
    @Override
    protected void put(String directory, String fileName, File file) throws NoSuchAlgorithmException, IOException, CloudException {
        logger.debug("enter - put(String, String, File)");
        try {
            BlobStoreContext ctx = cloud.getSwiftContext();

            try {
                org.jclouds.blobstore.BlobStore client = getClient(ctx);
                Blob blob = client.newBlob(fileName);

                blob.setPayload(file);
                blob.getMetadata().getContentMetadata().setContentLength(file.length());
                //blob.generateMD5();
                client.putBlob(toSwift(directory), blob);
            }
            finally {
                ctx.close();
            }
        }
        finally {
            logger.debug("exit - put(String, String, File)");
        }
    }
    
    @Override
    protected void put(String directory, String fileName, String content) throws NoSuchAlgorithmException, IOException, CloudException {
        logger.debug("enter - put(String, String, String)");
        try {
            File tmp = File.createTempFile(fileName, ".txt");
            PrintWriter writer;
            
            try {
                writer = new PrintWriter(new OutputStreamWriter(new FileOutputStream(tmp)));
                writer.print(content);
                writer.flush();
                writer.close();
                put(directory, fileName, tmp);
            }
            finally {
                tmp.delete();
            }
        }
        finally {
            logger.debug("exit - put(String, String, String)");
        }
    }

    @Override
    public boolean isPublic(String directory, String file) throws CloudException, InternalException {
        return false;
    }
    
    
    @Override
    protected String verifyName(String name, boolean container) throws CloudException {
        if( name == null ) {
            return null;
        }
        name = name.toLowerCase().trim();
        if( name.length() > 255 ) {
            String extra = name.substring(255);
            int idx = extra.indexOf(".");
            
            if( idx > -1 ) {
                throw new CloudException("Azure names are limited to 255 characters.");
            }
            name = name.substring(0,255);
        }
        while( name.indexOf("--") != -1 ) {
            name = name.replaceAll("--", "-");         
        }
        StringBuilder str = new StringBuilder();

        for( int i=0; i<name.length(); i++ ) {
            char c = name.charAt(i);
            
            if( Character.isLetterOrDigit(c) || c == '-' ) {
                if( container ) {
                    if( i == 0 && !Character.isLetter(c) ) {
                        throw new CloudException("Azure container names must start with a letter.");
                    }
                }
                str.append(c);
            }
            else if( c == '.' ) {
                str.append(c);
            }
        }
        name = str.toString();
        
        if( name.length() < 1 ) { 
            return "000";
        }
        while( name.charAt(name.length()-1) == '-' ) {
            name = name.substring(0,name.length()-1);
            if( name.length() < 1 ) { 
                return "000";
            }
        }
        if( name.length() < 1 ) { 
            return "000";
        }
        else if( name.length() == 1 ) {
            name = name + "00";
        }
        else if ( name.length() == 2 ) {
            name = name + "0";
        }
        return name;
    }
}
